import logging

import aiohttp
import orjson
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

from .errors import AporiaError
from .logging_utils import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)


QUERY_MAX_ATTEMPTS = 4
QUERY_TIMEOUT = 10
QUERY_RETRY_INITIAL_SLEEP = 2
DEFAULT_TIMEOUT_SEC = 30
HTTP_502_VERBOSE_ERROR = """The server couldn't handle the rate at which you are sending requests.
Please try one of the following solutions:
    * Reduce the amount of log_* calls per second
    * Upgrade your cluster so it can support a higher load.
"""


class GraphQLClient:
    """Asynchronous graphql client."""

    def __init__(self, token: str, host: str, port: int):
        """Initialize a GraphQLClient instance.

        Args:
            token (str): Authorization token
            host (str): Controller address
            port (int): Controller port
        """
        logger.debug("Initializing GraphQL client.")
        if not host.startswith("http"):
            host = "https://{}".format(host)

        self.request_url = "{}:{}/v1/controller/graphql".format(host, port)

        self.headers = {
            "Authorization": "Bearer {}".format(token),
            "Content-Type": "application/json",
        }
        self.session = None

    async def open(self):
        """Opens the HTTP session."""
        logger.debug("Creating HTTP session with controller.")
        self.session = aiohttp.ClientSession(headers=self.headers)

    @retry(
        stop=stop_after_attempt(QUERY_MAX_ATTEMPTS),
        wait=wait_exponential(min=QUERY_RETRY_INITIAL_SLEEP),
        retry=retry_if_exception(lambda err: not isinstance(err, AporiaError)),
        reraise=True,
    )
    async def query_with_retries(
        self, query: str, variables: dict, timeout: int = DEFAULT_TIMEOUT_SEC
    ) -> dict:
        """Executes a GraphQL query with retries in case of failure.

        Args:
            query (str): GraphQL query string
            variables (dict): Variables for the query
            timeout (int): Timeout for the entire request, in seconds. Defaults to 5 minutes.

        Returns:
            dict: GraphQL query result
        """
        return await self.query(query, variables, timeout)

    async def query(self, query: str, variables: dict, timeout: int = DEFAULT_TIMEOUT_SEC) -> dict:
        """Executes a GraphQL query and returns the result.

        Args:
            query (str): GraphQL query string
            variables (dict): Variables for the query
            timeout (int): Timeout for the entire request, in seconds. Defaults to 5 minutes.

        Returns:
            dict: GraphQL query result
        """
        logger.debug("Sending GraphQL query: {}, variables: {}".format(query, variables))

        data = self._serialize_request(query, variables)
        # Note: By default, aiohttp uses SSL and verifies the certificate in HTTPS requests
        async with self.session.post(  # type: ignore
            url=self.request_url,
            data=data,
            timeout=aiohttp.ClientTimeout(total=timeout),
        ) as response:

            if response.status != 200:
                if response.status == 400:
                    errors = (await response.json())["errors"]  # type: ignore
                    raise AporiaError(
                        short_message=errors[0]["message"],
                        verbose_message=errors[0].get("extensions", {}).get("verbose_message"),
                    )

                elif response.status == 401:
                    raise AporiaError("Authentication failed, please check your token.")

                elif response.status == 502:
                    raise AporiaError(
                        short_message="HTTP Gateway received invalid response from server",
                        verbose_message=HTTP_502_VERBOSE_ERROR,
                    )
                else:
                    raise AporiaError("Unexpected HTTP error {}".format(response.status))

            return (await response.json())["data"]  # type: ignore

    @staticmethod
    def _serialize_request(query: str, variables: dict) -> bytes:
        request = {"query": query, "variables": variables}
        try:
            if hasattr(orjson, "OPT_SERIALIZE_NUMPY"):
                data = orjson.dumps(request, option=orjson.OPT_SERIALIZE_NUMPY)  # type: ignore
            else:
                data = orjson.dumps(request)

            return data
        except TypeError as err:
            raise AporiaError(str(err))

    async def close(self):
        """Closes the http session."""
        await self.session.close()



import logging
import sys

LOGGER_NAME = "aporia"
LOG_FORMAT = "{asctime}:{levelname}:{name}:{message}"
DEFAULT_LOG_LEVEL = logging.INFO

logger = logging.getLogger(LOGGER_NAME)


def init_logger():
    """Initializes the Aporia logger."""
    logger.setLevel(DEFAULT_LOG_LEVEL)

    if not logger.hasHandlers():
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT, style="{"))
        logger.addHandler(handler)


def set_log_level(debug: bool):
    """Sets the log level of the Aporia logger."""
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(DEFAULT_LOG_LEVEL)



import logging
from typing import Dict, List, Optional

from aporia.core.errors import AporiaError
from aporia.core.graphql_client import GraphQLClient
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldCategory, FieldType

VALID_FIELD_TYPES = [field_type.value for field_type in FieldType]

logger = logging.getLogger(LOGGER_NAME)


async def run_create_model_version_query(
    graphql_client: GraphQLClient,
    model_id: str,
    model_version: str,
    model_type: str,
    features: Dict[str, str],
    prediction: Dict[str, str],
    raw_inputs: Optional[Dict[str, str]] = None,
    outputs: Optional[Dict[str, str]] = None,
    metrics: Optional[Dict[str, str]] = None,
):
    """Defines the schema for a specific model version.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model Version
        model_type (str): Model type
        features (Dict[str, str]): Feature fields
        prediction (Dict[str, str]): Prediction fields
        raw_inputs (Dict[str, str], optional): Raw input fields. Defaults to None.
        outputs (Dict[str, str], optional): Output fields. Defaults to None.
        metrics (Dict[str, str], optional): Prediction metric fields. Defaults to None.
    """
    query = """
        mutation CreateModelVersion(
            $modelId: String!,
            $modelVersion: String!,
            $modelType: String!,
            $features: [Field]!,
            $prediction: [Field]!,
            $rawInputs: [Field],
            $outputs: [Field],
            $metrics: [Field]
        ) {
            createModelVersion(
                modelId: $modelId,
                modelVersion: $modelVersion,
                modelType: $modelType,
                features: $features,
                prediction: $prediction
                rawInputs: $rawInputs,
                outputs: $outputs,
                metrics: $metrics
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "modelType": model_type,
        "features": prepare_fields(features),
        "prediction": prepare_fields(prediction),
        "rawInputs": None if raw_inputs is None else prepare_fields(raw_inputs),
        "outputs": None if outputs is None else prepare_fields(outputs),
        "metrics": None if metrics is None else prepare_fields(metrics),
    }

    result = await graphql_client.query_with_retries(query, variables)
    for warning in result["createModelVersion"]["warnings"]:
        logger.warning(warning)


def prepare_fields(fields: Dict[str, str]) -> List[Dict[str, str]]:
    """Creates a list of Field GraphQL objects from a fields dict.

    Args:
        fields (Dict[str, str]): Fields dict

    Returns:
        List[Dict[str, str]]: List of GraphQL Field objects.
    """
    return [{"name": field_name, "type": field_type} for field_name, field_type in fields.items()]


def validate_fields_input(
    features: Dict[str, str],
    prediction: Dict[str, str],
    raw_inputs: Optional[Dict[str, str]] = None,
    outputs: Optional[Dict[str, str]] = None,
    metrics: Optional[Dict[str, str]] = None,
):
    """Checks if the fields passed to create_model_version are valid.

    Args:
        prediction (Dict[str, str]): Prediction fields
        features (Dict[str, str]): Feature fields
        raw_inputs (Dict[str, str], optional): Raw input fields. Defaults to None.
        outputs (Dict[str, str], optional): Output fields. Defaults to None.
        metrics (Dict[str, str], optional): Prediction metric fields. Defaults to None.
    """
    fields_to_validate = {FieldCategory.FEATURES: features, FieldCategory.PREDICTION: prediction}

    if raw_inputs is not None:
        fields_to_validate[FieldCategory.RAW_INPUTS] = raw_inputs

    if outputs is not None:
        fields_to_validate[FieldCategory.OUTPUTS] = outputs

    if metrics is not None:
        fields_to_validate[FieldCategory.METRICS] = metrics

    for category, fields in fields_to_validate.items():
        if not isinstance(fields, dict):
            raise AporiaError("{} parameter must be a dict".format(category.value))

        if len(fields) == 0:
            raise AporiaError("{} parameter must contain items".format(category.value))

        for key, value in fields.items():
            if not isinstance(key, str):
                raise AporiaError(
                    "Invalid field name {} in the {} parameter - field names must be strings".format(
                        key, category.value
                    )
                )

            if value not in VALID_FIELD_TYPES:
                raise AporiaError(
                    "Invalid field type {} in the {} parameter - valid field types are {}".format(
                        value, category.value, VALID_FIELD_TYPES
                    )
                )



from typing import Dict

from aporia.core.graphql_client import GraphQLClient
from aporia.core.types.field import FieldCategory, FieldType


async def get_model_version(
    graphql_client: GraphQLClient, model_id: str, model_version: str
) -> Dict[FieldCategory, Dict[str, FieldType]]:
    """Fetches model version schema.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        model_id (str): Model ID
        model_version (str): Model version

    Notes:
        * This only fetches field categories that can be used
          when reporting training data (features, prediction, raw_inputs)

    Returns:
        Dict[FieldCategory, Dict[str, FieldType]]: Model version schema
    """
    query = """
        query GetModelVersion(
            $modelId: String!,
            $modelVersion: String!,
        ) {
            modelVersionSchema(
                modelId: $modelId,
                modelVersion: $modelVersion,
            ) {
                features {
                    name
                    type
                }
                prediction {
                    name
                    type
                }
                rawInputs {
                    name
                    type
                }
            }
        }
    """

    variables = {"modelId": model_id, "modelVersion": model_version}

    result = await graphql_client.query_with_retries(query, variables)
    return _build_model_version_schema(result["modelVersionSchema"])


def _build_model_version_schema(
    model_version_data: dict,
) -> Dict[FieldCategory, Dict[str, FieldType]]:
    schema = {}
    for category, fields in model_version_data.items():
        if fields is not None:
            schema[FieldCategory.from_camel_case(category)] = {
                field["name"]: FieldType(field["type"]) for field in fields
            }

    return schema



from contextlib import contextmanager
from functools import wraps
import logging
from typing import Callable

from .context import get_context
from .errors import AporiaError, handle_error
from .logging_utils import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)


def validate_model_ready(func: Callable) -> Callable:
    """Decorator that checks if the model is ready before calling the wrapped function.

    Args:
        func (Callable): Function to wrap

    Returns:
        Callable: Wrapped function
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):  # noqa: ANN001, ANN201
        if not getattr(self, "_model_ready", False):
            return

        return func(self, *args, **kwargs)

    return wrapper


class BaseModel:
    """Base class for model objects."""

    def __init__(self, model_id: str, model_version: str, set_ready_when_done: bool = True):
        """Initializes a BaseModel object.

        Args:
            model_id (str): Model identifier, as received from the Aporia dashboard.
            model_version (str): Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
            set_ready_when_done (bool): True if self._model_ready should be set to
                True when initialization is finished. False if the subclass wants to
                handle this. Defaults to True.
        """
        self._model_ready = False

        logger.debug(
            "Initializing model object for model {} version {}".format(model_id, model_version)
        )
        context = get_context()

        self.model_id = model_id
        self.model_version = model_version
        self._event_loop = context.event_loop
        self._graphql_client = context.graphql_client
        self._config = context.config

        if len(model_id) == 0 or len(model_version) == 0:
            raise AporiaError("model_id and model_version must be non-empty strings")

        if set_ready_when_done:
            self._model_ready = True

    @contextmanager
    def _handle_error(self, message_format: str):
        try:
            yield
        except Exception as err:
            handle_error(
                message_format=message_format,
                verbose=self._config.verbose,
                throw_errors=self._config.throw_errors,
                debug=self._config.debug,
                original_exception=err,
                log_level=logging.ERROR,
            )



from .config import Config
from .errors import AporiaError
from .event_loop import EventLoop
from .graphql_client import GraphQLClient


global_context = None


class Context:
    """Global context."""

    def __init__(
        self,
        graphql_client: GraphQLClient,
        event_loop: EventLoop,
        config: Config,
    ):
        """Initializes the context.

        Args:
            graphql_client (GraphQLClient): GraphQL client.
            event_loop (EventLoop): Event loop.
            config (Config): Aporia config.
        """
        self.graphql_client = graphql_client
        self.event_loop = event_loop
        self.config = config

        self._open_graphql_client()

    def _open_graphql_client(self):
        self.event_loop.run_coroutine(self.graphql_client.open())

    def shutdown(self):
        """Cleans up objects managed by the context."""
        self.event_loop.run_coroutine(self.graphql_client.close())


def init_context(graphql_client: GraphQLClient, event_loop: EventLoop, config: Config):
    """Initializes the global context.

    Args:
        graphql_client (GraphQLClient): GraphQL client
        event_loop (EventLoop): Event loop
        config (Config): Configuration
    """
    global global_context

    global_context = Context(graphql_client=graphql_client, event_loop=event_loop, config=config)


def reset_context():
    """Resets the global context."""
    global global_context
    if global_context is None:
        raise AporiaError("Aporia was not initialized.")

    global_context.shutdown()
    global_context = None


def get_context() -> Context:
    """Returns the global context."""
    if global_context is None:
        raise AporiaError("Aporia was not initialized.")

    return global_context



import logging
from typing import Optional

from .logging_utils import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)


class AporiaError(Exception):
    """Base class for Aporia SDK exceptions."""

    def __init__(self, short_message: str, verbose_message: Optional[str] = None):
        """Initializes an AporiaError.

        Args:
            short_message: Short error message
            verbose_message: Verbose error message. Defaults to the short message.
        """
        super().__init__(short_message)

        self.short_message = short_message
        self.verbose_message = short_message if verbose_message is None else verbose_message


def handle_error(
    message_format: str,
    verbose: bool,
    throw_errors: bool,
    debug: bool,
    log_level: int = logging.ERROR,
    original_exception: Optional[Exception] = None,
):
    """Handles an error with either a log or an exception.

    Args:
        message_format: Error message.
        verbose: True to display verbose error messages, False otherwise.
        throw_errors: True if exceptions should be raised
        debug: True if stack trace should be added to the log or exception.
        log_level: Log level for log messages. Defaults to logging.ERROR.
        original_exception: Original exception for stack trace. Defaults to None.
    """
    formatted_message = message_format
    if original_exception is not None:
        if verbose and isinstance(original_exception, AporiaError):
            formatted_message = message_format.format(original_exception.verbose_message)
        else:
            formatted_message = message_format.format(str(original_exception))

    if throw_errors:
        raise AporiaError(formatted_message) from original_exception
    else:
        logger.log(level=log_level, msg=formatted_message, exc_info=debug)



from datetime import datetime
from enum import Enum
from typing import Union


FieldValue = Union[float, int, str, bool, datetime]


class FieldType(Enum):
    """Field types."""

    NUMERIC = "numeric"
    CATEGORICAL = "categorical"
    STRING = "string"
    BOOLEAN = "boolean"
    DATETIME = "datetime"


class FieldCategory(Enum):
    """Field categories."""

    FEATURES = "features"
    PREDICTION = "prediction"
    METRICS = "metrics"
    RAW_INPUTS = "raw_inputs"
    OUTPUTS = "outputs"

    @staticmethod
    def from_camel_case(value: str) -> "FieldCategory":
        """Builds a field cateogry from a camelCase string.

        Args:
            value (str): Category value

        Returns:
            FieldCategory: Category that matches the given value
        """
        if value == "rawInputs":
            return FieldCategory.RAW_INPUTS

        return FieldCategory(value)
